from enum import IntEnum

import attr
import numpy as np

from nlisim.cell import CellData, CellList
from nlisim.coordinates import Point, Voxel
from nlisim.module import Module, ModuleState
from nlisim.modules.geometry import TissueTypes
from nlisim.random import rg
from nlisim.state import State


class FungusCellData(CellData):
    class Status(IntEnum):
        DRIFTING = 99
        RESTING = 0  # CONIDIA relevant
        SWOLLEN = 1  # CONIDIA relevant
        GERMINATED = 2  # CONIDIA relevant
        GROWABLE = 3  # HYPHAE relevant
        GROWN = 4  # HYPHAE relevant
        DEAD = 5

    class Form(IntEnum):
        CONIDIA = 0
        HYPHAE = 1

    FUNGUS_FIELDS = [
        ('form', 'u1'),
        ('status', 'u1'),
        ('iteration', 'i4'),
        ('mobile', 'b1'),
        ('internalized', 'b1'),
        ('iron', 'f8'),
        ('health', 'f8'),
    ]

    dtype = np.dtype(CellData.FIELDS + FUNGUS_FIELDS, align=True)  # type: ignore

    @classmethod
    def create_cell_tuple(
        cls,
        *,
        iron: float = 0,
        status: Status = Status.RESTING,
        form: Form = Form.CONIDIA,
        iteration=0,
        mobile=False,
        internalized=False,
        health=100,
        **kwargs,
    ) -> np.record:

        return CellData.create_cell_tuple(**kwargs) + (
            form,
            status,
            iteration,
            mobile,
            internalized,
            iron,
            health,
        )


@attr.s(kw_only=True, frozen=True, repr=False)
class FungusCellList(CellList):
    CellDataClass = FungusCellData

    def iron_uptake(self, iron: np.ndarray, iron_max: float, iron_min: float, iron_absorb: float):
        """Absorb iron from external environment."""
        cells = self.cell_data
        for vox_index in np.argwhere(iron > iron_min):
            vox = Voxel(x=vox_index[2], y=vox_index[1], z=vox_index[0])

            cells_here = self.get_cells_in_voxel(vox)

            indices = []
            for index in cells_here:
                if (
                    cells[index]['form'] == FungusCellData.Form.HYPHAE.value
                    and np.invert(cells[index]['internalized'])
                    and cells[index]['iron'] < iron_max
                ):
                    indices.append(index)

            if len(indices) > 0:
                iron_split = iron_absorb * (iron[vox.z, vox.y, vox.x] / len(indices))
                for cell_index in indices:
                    cells[cell_index]['iron'] += iron_split
                    if cells[cell_index]['iron'] > iron_max:
                        cells[cell_index]['iron'] = iron_max

                iron[vox.z, vox.y, vox.x] = (1 - iron_absorb) * iron[vox.z, vox.y, vox.x]

    def spawn_hypahael_cell(self, children):
        children['status'] = FungusCellData.Status.GROWABLE
        children['form'] = FungusCellData.Form.HYPHAE
        children['mobile'] = False
        self.extend(children)

    def spawn_spores(self, points):
        n, m = points.shape

        if m != 3:
            raise ValueError('Invalid shape for a point object')

        spores = FungusCellData(n, initialize=True)
        spores['point'] = points
        spores['status'] = FungusCellData.Status.RESTING

        self.extend(spores)

    def initialize_spores(self, tissue: np.ndarray, init_num: int):
        """Initialize spores on epithelium cells."""
        grid = self.grid
        if init_num > 0:
            points = np.zeros((init_num, 3))
            indices = np.argwhere(tissue == TissueTypes.EPITHELIUM.value)
            if len(indices) > 0:
                rg.shuffle(indices)
                for i in range(init_num):
                    x = rg.uniform(grid.xv[indices[i][2]], grid.xv[indices[i][2] + 1])
                    y = rg.uniform(grid.yv[indices[i][1]], grid.yv[indices[i][1] + 1])
                    z = rg.uniform(grid.zv[indices[i][0]], grid.zv[indices[i][0] + 1])

                    point = Point(x=x, y=y, z=z)
                    points[i] = point

                self.spawn_spores(points)

    def grow_hyphae(self, iron_min_grow, grow_time, p_branch, spacing):
        """Grow fungal hyphae."""
        cells = self.cell_data

        conidia_indices = self.alive(
            (cells['form'] == FungusCellData.Form.CONIDIA)
            & (cells['status'] == FungusCellData.Status.GERMINATED)
            & (np.invert(cells['internalized']))
        )

        hyphae_indices = self.alive(
            (cells['form'] == FungusCellData.Form.HYPHAE)
            & (cells['status'] == FungusCellData.Status.GROWABLE)
            & (np.invert(cells['internalized']))
            & (cells['iron'] > iron_min_grow)
            & (cells['iteration'] > grow_time)
        )

        # grow conidia
        if len(conidia_indices) != 0:
            cells['status'][conidia_indices] = FungusCellData.Status.GROWN
            cells['form'][conidia_indices] = FungusCellData.Form.HYPHAE
            children = FungusCellData(len(conidia_indices), initialize=True)
            children['iron'] = cells['iron'][conidia_indices]
            growth = spacing * (rg.random((len(conidia_indices), 3)) * 2 - 1)
            children['point'] = cells['point'][conidia_indices] + growth
            self.spawn_hypahael_cell(children)

        # grow hyphae
        if len(hyphae_indices) != 0:
            cells['status'][hyphae_indices] = FungusCellData.Status.GROWN
            branch_mask = rg.random(len(hyphae_indices)) < p_branch
            not_branch_indices = (np.invert(branch_mask)).nonzero()[0]
            branch_indices = branch_mask.nonzero()[0]

            elongate_children = FungusCellData(len(hyphae_indices), initialize=True)
            branch_children = FungusCellData(len(branch_indices), initialize=True)

            elongate_children['iron'] = cells['iron'][hyphae_indices] / 2
            growth = spacing * (rg.random((len(hyphae_indices), 3)) * 2 - 1)
            elongate_children['point'] = cells['point'][hyphae_indices] + growth

            if len(branch_indices) != 0:
                elongate_children['iron'][branch_indices] = (
                    cells['iron'][hyphae_indices[branch_indices]] / 3
                )

                branch_children['iron'] = cells['iron'][hyphae_indices[branch_indices]] / 3
                growth = spacing * (rg.random((len(hyphae_indices[branch_indices]), 3)) * 2 - 1)
                branch_children['point'] = cells['point'][hyphae_indices[branch_indices]] + growth

            # update iron in orignal cells
            cells['iron'][hyphae_indices[not_branch_indices]] /= 2
            cells['iron'][hyphae_indices[branch_indices]] /= 3

            self.spawn_hypahael_cell(elongate_children)
            self.spawn_hypahael_cell(branch_children)

    def age(self):
        """Add one iteration to all alive cells."""
        np.add.at(self.cell_data['iteration'], self.alive(), 1)

    def kill(self):
        """If a cell have 0 health point or out of range, kill the cell."""
        cells = self.cell_data
        mask = cells.point_mask(cells['point'], self.grid)
        indices = self.alive(np.logical_or(cells['health'] <= 0, np.invert(mask)))

        cells['status'][indices] = FungusCellData.Status.DEAD
        cells['dead'][indices] = True

    def change_status(self, p_internal_swell: float, rest_time: int, swell_time: int):
        cells = self.cell_data

        indices = self.alive(
            cells['form'] != FungusCellData.Form.HYPHAE,
        )

        internalized_indices = (cells['internalized'][indices]).nonzero()[0]
        not_internalized_indices = (np.invert(cells['internalized'][indices])).nonzero()[0]

        internalized_rest_indices = np.logical_and(
            cells['status'][internalized_indices] == FungusCellData.Status.RESTING,
            cells['iteration'][internalized_indices] >= rest_time,
        ).nonzero()[0]

        internalized_swollen_indices = np.logical_and(
            cells['status'][internalized_indices] == FungusCellData.Status.SWOLLEN,
            cells['iteration'][internalized_indices] >= swell_time,
        ).nonzero()[0]

        # internal fungus with REST status
        swall_mask = rg.random(len(internalized_rest_indices)) < p_internal_swell
        internalized_rest_indices = swall_mask.nonzero()[0]

        cells['status'][internalized_rest_indices] = FungusCellData.Status.SWOLLEN
        cells['iteration'][internalized_rest_indices] = 0

        # internal fungus with SWOLLEN status
        cells['status'][internalized_swollen_indices] = FungusCellData.Status.GERMINATED
        cells['iteration'][internalized_swollen_indices] = 0

        rest_indices = np.logical_and(
            cells['status'][not_internalized_indices] == FungusCellData.Status.RESTING,
            cells['iteration'][not_internalized_indices] >= rest_time,
        ).nonzero()[0]
        swollen_indices = np.logical_and(
            cells['status'][not_internalized_indices] == FungusCellData.Status.SWOLLEN,
            cells['iteration'][not_internalized_indices] >= swell_time,
        ).nonzero()[0]

        # free fungus with REST status
        cells['status'][rest_indices] = FungusCellData.Status.SWOLLEN
        cells['iteration'][rest_indices] = 0

        # free fungus with SWOLLEN status
        cells['status'][swollen_indices] = FungusCellData.Status.GERMINATED
        cells['iteration'][swollen_indices] = 0


def cell_list_factory(self: 'FungusState'):
    return FungusCellList(grid=self.global_state.grid)


@attr.s(kw_only=True)
class FungusState(ModuleState):
    cells: FungusCellList = attr.ib(default=attr.Factory(cell_list_factory, takes_self=True))
    # init_num: int = 0
    # p_lodge: float = 0
    # p_internal_swell: float = 0.05
    # iron_min: int = 0
    # iron_max: float = 0.0
    # iron_absorb: float = 0.0
    # spacing: float = 0.0
    # iron_min_grow: float = 0.0
    # grow_time: float = 0.0
    # p_branch: float = 0.0
    # p_internalize: float = 0.0
    health: float = 100.0


class Fungus(Module):
    name = 'fungus'
    StateClass = FungusState

    # defaults config opt
    defaults = {
        'init_num': '10',
        'p_lodge': '0.1',
        'p_internal_swell': '0.05',
        'iron_min': '10',
        'iron_max': '20',
        'iron_absorb': '1',
        'spacing': '0.1',
        'iron_min_grow': '5',
        'p_branch': '0.2',
        'p_internalize': '0.1',
        'health': '100.0',
        'rest_time': '1',
        'swell_time': '1',
        'grow_time': '1',
        # ...
    }

    def initialize(self, state: State):
        fungus: FungusState = state.fungus
        # grid: RectangularGrid = state.grid
        tissue = state.geometry.lung_tissue

        self.init_num = self.config.getint('init_num')
        self.p_lodge = self.config.getfloat('p_lodge')
        self.p_internal_swell = self.config.getfloat('p_internal_swell')
        self.iron_min = self.config.getint('iron_min')
        self.iron_max = self.config.getfloat('iron_max')
        self.iron_absorb = self.config.getfloat('iron_absorb')
        self.spacing = self.config.getfloat('spacing')
        self.iron_min_grow = self.config.getfloat('iron_min_grow')
        self.p_branch = self.config.getfloat('p_branch')
        self.p_internalize = self.config.getfloat('p_internalize')
        self.rest_time = self.config.getint('rest_time')
        self.swell_time = self.config.getint('swell_time')
        self.grow_time = self.config.getint('grow_time')

        fungus.health = self.config.getfloat('health')

        cells = fungus.cells
        cells.initialize_spores(tissue, self.init_num)

        return state

    def advance(self, state: State, previous_time: float):
        cells = state.fungus.cells

        cells.kill()  # clear dead cell
        cells.age()
        cells.change_status(self.p_internal_swell, self.rest_time, self.swell_time)
        if hasattr(state, 'molecules'):
            iron = state.molecules.grid['iron']
            cells.iron_uptake(iron, self.iron_max, self.iron_min, self.iron_absorb)
        cells.grow_hyphae(self.iron_min_grow, self.grow_time, self.p_branch, self.spacing)

        return state



import numpy as np
from pytest import fixture

from nlisim.coordinates import Point
from nlisim.grid import RectangularGrid
from nlisim.modules.epithelium import (
    EpitheliumCellData,
    EpitheliumCellList,
)
from nlisim.modules.fungus import (
    FungusCellData,
    FungusCellList,
)


@fixture
def iron():
    # a 10 x 10 x 10 grid with 10 iron
    i = np.empty((10, 10, 10))
    i.fill(10)
    yield i


@fixture
def n_cyto():
    # a 10 x 10 x 10 grid with 10 iron
    i = np.empty((10, 10, 10))
    i.fill(0)
    yield i


@fixture
def m_cyto():
    # a 10 x 10 x 10 grid with 10 iron
    i = np.empty((10, 10, 10))
    i.fill(0)
    yield i


@fixture
def tissue():
    # a 10 x 10 x 10 grid of blood
    t = np.empty((10, 10, 10))
    t.fill(3)
    yield t


@fixture
def epithelium_list(grid: RectangularGrid):
    epithelium = EpitheliumCellList(grid=grid)
    yield epithelium


@fixture
def populated_epithelium(epithelium_list: EpitheliumCellList, grid: RectangularGrid):
    epithelium_list.append(
        EpitheliumCellData.create_cell(
            point=Point(x=grid.x[3], y=grid.y[3], z=grid.z[3]),
        )
    )

    yield epithelium_list


@fixture
def fungus_list(grid: RectangularGrid):
    fungus = FungusCellList(grid=grid)
    yield fungus


@fixture
def populated_fungus(fungus_list: FungusCellList, grid: RectangularGrid):
    points = []
    for i in range(int(grid.x[1]), int(grid.x[6]), 10):
        points.append(Point(x=i, y=i, z=i))

    for point in points:
        fungus_list.append(
            FungusCellData.create_cell(
                point=point,
                status=FungusCellData.Status.GROWABLE,
                form=FungusCellData.Form.HYPHAE,
                iron=0,
                mobile=False,
            )
        )
    yield fungus_list


# tests

# internalize_conidia
def test_internalize_conidia_none(
    populated_epithelium: EpitheliumCellList,
    grid: RectangularGrid,
    fungus_list: FungusCellList,
):
    cell = populated_epithelium[0]
    vox = grid.get_voxel(cell['point'])
    assert len(fungus_list.get_cells_in_voxel(vox)) == 0

    populated_epithelium.internalize_conidia(0, 10, 1, grid, fungus_list)

    assert populated_epithelium.len_phagosome(0) == 0
    for v in cell['phagosome']:
        assert v == -1


def test_internalize_conidia_1(
    epithelium_list: EpitheliumCellList, grid: RectangularGrid, fungus_list: FungusCellList
):
    point = Point(x=35, y=35, z=35)
    epithelium_list.append(EpitheliumCellData.create_cell(point=point))
    fungus_list.append(
        FungusCellData.create_cell(point=point, status=FungusCellData.Status.RESTING)
    )

    vox = grid.get_voxel(epithelium_list[0]['point'])

    epithelium_list.internalize_conidia(0, 10, 1, grid, fungus_list)

    assert grid.get_voxel(fungus_list[0]['point']) == vox
    assert epithelium_list.len_phagosome(0) == 1
    assert 0 in epithelium_list[0]['phagosome']


def test_internalize_conidia_2(
    epithelium_list: EpitheliumCellList, grid: RectangularGrid, fungus_list: FungusCellList
):
    point = Point(x=35, y=35, z=35)
    epithelium_list.append(EpitheliumCellData.create_cell(point=point))
    fungus_list.append(
        FungusCellData.create_cell(point=point, status=FungusCellData.Status.RESTING)
    )
    fungus_list.append(
        FungusCellData.create_cell(point=point, status=FungusCellData.Status.RESTING)
    )

    vox = grid.get_voxel(epithelium_list[0]['point'])

    epithelium_list.internalize_conidia(0, 10, 1, grid, fungus_list)

    assert grid.get_voxel(fungus_list[0]['point']) == vox
    assert epithelium_list.len_phagosome(0) == 2
    assert 0 in epithelium_list[0]['phagosome']
    assert 1 in epithelium_list[0]['phagosome']


def test_internalize_conidia_2b(
    epithelium_list: EpitheliumCellList, grid: RectangularGrid, fungus_list: FungusCellList
):
    point = Point(x=35, y=35, z=35)
    epithelium_list.append(EpitheliumCellData.create_cell(point=point))
    fungus_list.append(
        FungusCellData.create_cell(point=point, status=FungusCellData.Status.RESTING)
    )
    fungus_list.append(
        FungusCellData.create_cell(point=point, status=FungusCellData.Status.RESTING)
    )

    vox = grid.get_voxel(epithelium_list[0]['point'])
    fungus_list[0]['internalized'] = True  # say by macrophage

    epithelium_list.internalize_conidia(0, 10, 1, grid, fungus_list)

    assert grid.get_voxel(fungus_list[0]['point']) == vox
    assert epithelium_list.len_phagosome(0) == 1
    assert 0 not in epithelium_list[0]['phagosome']
    assert 1 in epithelium_list[0]['phagosome']


def test_internalize_conidia_max(
    epithelium_list: EpitheliumCellList, grid: RectangularGrid, fungus_list: FungusCellList
):
    point = Point(x=35, y=35, z=35)
    epithelium_list.append(EpitheliumCellData.create_cell(point=point))
    fungus_list.append(
        FungusCellData.create_cell(point=point, status=FungusCellData.Status.RESTING)
    )

    max_spores = 10
    epithelium_list[0]['phagosome'][:max_spores] = 99  # artificially fill

    epithelium_list.internalize_conidia(0, max_spores, 1, grid, fungus_list)

    assert epithelium_list.len_phagosome(0) == max_spores
    assert 0 not in epithelium_list[0]['phagosome']
    assert not fungus_list[0]['internalized']


def test_dead_conidia_1(
    epithelium_list: EpitheliumCellList, grid: RectangularGrid, fungus_list: FungusCellList
):
    point = Point(x=35, y=35, z=35)
    epithelium_list.append(EpitheliumCellData.create_cell(point=point))
    fungus_list.append(
        FungusCellData.create_cell(point=point, status=FungusCellData.Status.RESTING)
    )

    epithelium_list.internalize_conidia(0, 10, 1, grid, fungus_list)
    fungus_list[0]['dead'] = True  # simulate killing

    epithelium_list.remove_dead_fungus(fungus_list, grid)

    assert epithelium_list.len_phagosome(0) == 0
    assert 0 not in epithelium_list[0]['phagosome']


def test_produce_cytokines_0(
    epithelium_list: EpitheliumCellList,
    grid: RectangularGrid,
    fungus_list: FungusCellList,
    m_cyto,
    n_cyto,
):
    s_det = 0
    h_det = 0
    cyto_rate = 10

    assert m_cyto[3, 3, 3] == 0
    assert n_cyto[3, 3, 3] == 0

    point = Point(x=35, y=35, z=35)
    epithelium_list.append(EpitheliumCellData.create_cell(point=point))

    fungus_list.append(
        FungusCellData.create_cell(
            point=point,
            status=FungusCellData.Status.RESTING,
            form=FungusCellData.Form.CONIDIA,
            iron=0,
            mobile=False,
        )
    )

    # fungus is not swollen or germinated
    epithelium_list.cytokine_update(s_det, h_det, cyto_rate, m_cyto, n_cyto, fungus_list, grid)

    assert m_cyto[3, 3, 3] == 0
    assert n_cyto[3, 3, 3] == 0

    fungus_list[0]['status'] = FungusCellData.Status.SWOLLEN

    epithelium_list.cytokine_update(s_det, h_det, cyto_rate, m_cyto, n_cyto, fungus_list, grid)

    assert m_cyto[3, 3, 3] == 10
    assert n_cyto[3, 3, 3] == 10


def test_produce_cytokines_0b(
    epithelium_list: EpitheliumCellList,
    grid: RectangularGrid,
    fungus_list: FungusCellList,
    m_cyto,
    n_cyto,
):
    s_det = 0
    h_det = 0
    cyto_rate = 10

    assert m_cyto[3, 3, 3] == 0
    assert n_cyto[3, 3, 3] == 0

    point = Point(x=35, y=35, z=35)
    epithelium_list.append(EpitheliumCellData.create_cell(point=point))

    fungus_list.append(
        FungusCellData.create_cell(
            point=point,
            status=FungusCellData.Status.SWOLLEN,
            form=FungusCellData.Form.CONIDIA,
            iron=0,
            mobile=False,
        )
    )

    epithelium_list.cytokine_update(s_det, h_det, cyto_rate, m_cyto, n_cyto, fungus_list, grid)

    assert m_cyto[3, 3, 3] == 10
    assert n_cyto[3, 3, 3] == 10

    fungus_list.append(
        FungusCellData.create_cell(
            point=point,
            status=FungusCellData.Status.GROWABLE,
            form=FungusCellData.Form.HYPHAE,
            iron=0,
            mobile=False,
        )
    )

    epithelium_list.cytokine_update(s_det, h_det, cyto_rate, m_cyto, n_cyto, fungus_list, grid)

    assert m_cyto[3, 3, 3] == 20
    assert n_cyto[3, 3, 3] == 30


def test_produce_cytokines_2(
    epithelium_list: EpitheliumCellList,
    grid: RectangularGrid,
    fungus_list: FungusCellList,
    m_cyto,
    n_cyto,
):
    s_det = 1
    h_det = 2
    cyto_rate = 10

    assert m_cyto[3, 3, 3] == 0
    assert n_cyto[3, 3, 3] == 0

    point = Point(x=35, y=35, z=35)
    epithelium_list.append(EpitheliumCellData.create_cell(point=point))

    spoint = Point(x=15, y=35, z=35)
    fungus_list.append(
        FungusCellData.create_cell(
            point=spoint,
            status=FungusCellData.Status.SWOLLEN,
            form=FungusCellData.Form.CONIDIA,
            iron=0,
            mobile=False,
        )
    )

    epithelium_list.cytokine_update(s_det, h_det, cyto_rate, m_cyto, n_cyto, fungus_list, grid)

    assert m_cyto[3, 3, 3] == 0
    assert n_cyto[3, 3, 3] == 0

    fungus_list.append(
        FungusCellData.create_cell(
            point=spoint,
            status=FungusCellData.Status.GROWABLE,
            form=FungusCellData.Form.HYPHAE,
            iron=0,
            mobile=False,
        )
    )

    epithelium_list.cytokine_update(s_det, h_det, cyto_rate, m_cyto, n_cyto, fungus_list, grid)

    assert m_cyto[3, 3, 3] == 0
    assert n_cyto[3, 3, 3] == 10


def test_damage_conidia(
    epithelium_list: EpitheliumCellList, grid: RectangularGrid, fungus_list: FungusCellList
):
    kill = 2
    t = 1
    health = 100

    point = Point(x=35, y=35, z=35)
    epithelium_list.append(EpitheliumCellData.create_cell(point=point))
    fungus_list.append(
        FungusCellData.create_cell(point=point, status=FungusCellData.Status.RESTING)
    )

    epithelium_list.internalize_conidia(0, 10, 1, grid, fungus_list)

    epithelium_list.damage(kill, t, health, fungus_list)

    assert fungus_list.cell_data['health'][0] == 50

    epithelium_list.damage(kill, t, health, fungus_list)

    assert fungus_list.cell_data['health'][0] == 0


def test_kill_epithelium(
    epithelium_list: EpitheliumCellList, grid: RectangularGrid, fungus_list: FungusCellList
):
    # should release all conidia

    point = Point(x=35, y=35, z=35)
    epithelium_list.append(EpitheliumCellData.create_cell(point=point))
    fungus_list.append(
        FungusCellData.create_cell(point=point, status=FungusCellData.Status.RESTING)
    )

    fungus_list.cell_data['internalized'][0] = True
    epithelium_list[0]['phagosome'][0] = 0  # internalized

    epithelium_list.die_by_germination(fungus_list)
    assert fungus_list.cell_data['internalized'][0]
    assert epithelium_list.len_phagosome(0) == 1

    fungus_list.cell_data['status'][0] = FungusCellData.Status.GERMINATED

    epithelium_list.die_by_germination(fungus_list)

    assert not fungus_list.cell_data['internalized'][0]
    assert epithelium_list.len_phagosome(0) == 0


def test_kill_epithelium_n(
    epithelium_list: EpitheliumCellList, grid: RectangularGrid, fungus_list: FungusCellList
):
    # should release all conidia

    point = Point(x=35, y=35, z=35)
    epithelium_list.append(EpitheliumCellData.create_cell(point=point))
    for _ in range(0, 10):
        fungus_list.append(
            FungusCellData.create_cell(point=point, status=FungusCellData.Status.RESTING)
        )

    epithelium_list.internalize_conidia(0, 10, 1, grid, fungus_list)

    epithelium_list.die_by_germination(fungus_list)
    for i in range(0, 10):
        assert fungus_list.cell_data['internalized'][i]
    assert epithelium_list.len_phagosome(0) == 10

    fungus_list.cell_data['status'][6] = FungusCellData.Status.GERMINATED

    epithelium_list.die_by_germination(fungus_list)

    for i in range(0, 10):
        assert not fungus_list.cell_data['internalized'][i]
    assert epithelium_list.len_phagosome(0) == 0



import numpy as np


class Coordinate(np.ndarray):
    """A base class for representing 3D vectors as numpy arrays."""

    def __new__(cls, *, x: float = 0, y: float = 0, z: float = 0) -> 'Coordinate':
        return np.asarray([(z, y, x)], dtype=cls.dtype).reshape((3,)).view(cls)

    def __repr__(self):
        return f'{self.__class__.__name__}({self.x}, {self.y}, {self.z})'

    @classmethod
    def from_array(cls, array: np.ndarray) -> 'Coordinate':
        """Generate a `Coordinate` vector from a 3-element array."""
        if array.shape != (3,):
            raise ValueError('Invalid shape for a coordinate object')
        return array.view(cls)

    @property
    def x(self) -> float:
        """Return the `x`-coordinate of the vector."""
        return self[2]

    @x.setter
    def x(self, value: float):
        self[2] = value

    @property
    def y(self) -> float:
        """Return the `y`-coordinate of the vector."""
        return self[1]

    @y.setter
    def y(self, value: float):
        self[1] = value

    @property
    def z(self) -> float:
        """Return the `z`-coordinate of the vector."""
        return self[0]

    @z.setter
    def z(self, value: float):
        self[0] = value

    def norm(self):
        """Return the euclidean norm of the vector."""
        return np.linalg.norm(self, ord=2)


class Point(Coordinate):
    """An array subclass representing a point or vector in 3D space."""

    dtype = np.dtype((np.dtype('<f8'), (3,)))


class Voxel(Coordinate):
    """An array subclass representing the coordinates of a voxel in a 3D array."""

    dtype = np.dtype((np.dtype('i4'), (3,)))

    def __hash__(self):
        return hash(','.join([str(i) for i in self]))

    def __eq__(self, other):
        value = super().__eq__(other)
        if isinstance(value, np.ndarray):
            value = bool(value.all())
        return value

    def __ne__(self, other):
        return not self.__eq__(other)



from pathlib import Path
from typing import Dict, Iterable, Tuple

import numpy as np  # type: ignore
from vtk import (  # type: ignore
    vtkPoints,
    vtkPolyData,
    vtkStructuredPoints,
    vtkXMLImageDataWriter,
    vtkXMLPolyDataWriter,
)
from vtk.util.numpy_support import numpy_to_vtk  # type: ignore

from nlisim.cell import CellList
from nlisim.grid import RectangularGrid
from nlisim.modules.geometry import GeometryState
from nlisim.state import State


def convert_cells_to_vtk(cells: CellList) -> vtkPolyData:
    cell_data = cells.cell_data
    fields = dict(cell_data.dtype.fields)
    fields.pop('point')

    points = vtkPoints()
    poly = vtkPolyData()
    poly.SetPoints(points)

    if not len(cell_data):
        return poly

    # vtk uses coordinate ordering x, y, z while we use z, y, x.
    points.SetData(numpy_to_vtk(np.flip(cell_data['point'], axis=1)))
    point_data = poly.GetPointData()

    for field, (dtype, _) in fields.items():
        data = cell_data[field]

        # numpy_to_vtk doesn't handle bool for some reason
        if dtype == np.dtype('bool'):
            data = data.astype(np.dtype('uint8'))

        try:
            scalar = numpy_to_vtk(data)
        except Exception:
            print(f'Unhandled data type in field {field}')
            continue

        scalar.SetName(field)
        point_data.AddArray(scalar)

    return poly


def create_vtk_volume(grid: RectangularGrid, geometry: GeometryState) -> vtkStructuredPoints:
    x = grid.x
    y = grid.y
    z = grid.z
    vtk_grid = vtkStructuredPoints()
    vtk_grid.SetDimensions(len(x), len(y), len(z))

    # In theory, the rectangular grid used in our code is more general than
    # than a vtkStructuredPoints object.  In practice, we always construct
    # a uniform grid, so we choose to use the more efficient vtk data structure.
    vtk_grid.SetOrigin(0, 0, 0)
    vtk_grid.SetSpacing(x[1] - x[0], y[1] - y[0], z[1] - z[0])

    point_data = vtk_grid.GetPointData()

    # transform color values to get around categorical interpolation issue in visualization
    tissue = geometry.lung_tissue
    tissue[tissue == 0] = 4
    point_data.SetScalars(numpy_to_vtk(tissue.ravel()))
    return vtk_grid


def generate_vtk_objects(state: State) -> Tuple[vtkStructuredPoints, Dict[str, vtkPolyData]]:
    volume = create_vtk_volume(state.grid, state.geometry)
    cells = {
        'spore': convert_cells_to_vtk(state.fungus.cells),
        'epithelium': convert_cells_to_vtk(state.epithelium.cells),
        'macrophage': convert_cells_to_vtk(state.macrophage.cells),
        'neutrophil': convert_cells_to_vtk(state.neutrophil.cells),
    }

    return volume, cells


def generate_vtk(state: State, postprocess_step_dir: Path):
    volume, cells = generate_vtk_objects(state)

    grid_writer = vtkXMLImageDataWriter()
    grid_writer.SetDataModeToBinary()
    grid_writer.SetFileName(str(postprocess_step_dir / 'geometry_001.vti'))
    grid_writer.SetInputData(volume)
    grid_writer.Write()

    cell_writer = vtkXMLPolyDataWriter()
    cell_writer.SetDataModeToBinary()
    for module, data in cells.items():
        cell_writer.SetFileName(str(postprocess_step_dir / f'{module}_001.vtp'))
        cell_writer.SetInputData(data)
        cell_writer.Write()


def process_output(state_files: Iterable[Path], postprocess_dir: Path) -> None:
    for state_file_index, state_file in enumerate(sorted(state_files)):
        state = State.load(state_file)

        postprocess_step_dir = postprocess_dir / ('%03i' % (state_file_index + 1))
        postprocess_step_dir.mkdir()
        generate_vtk(state, postprocess_step_dir)



import attr
import matplotlib.pyplot as plt
import numpy as np

from nlisim.module import Module, ModuleState
from nlisim.state import State

MAX_ARRAY_LENGTH = 1000


@attr.s(kw_only=True)
class PlotState(ModuleState):
    step_num: int = attr.ib(default=0)
    fungal_burdens: np.ndarray = attr.ib(factory=lambda: np.zeros(MAX_ARRAY_LENGTH, dtype=int))
    macrophage_counts: np.ndarray = attr.ib(factory=lambda: np.zeros(MAX_ARRAY_LENGTH, dtype=int))
    neutrophil_counts: np.ndarray = attr.ib(factory=lambda: np.zeros(MAX_ARRAY_LENGTH, dtype=int))
    time_steps: np.ndarray = attr.ib(factory=lambda: np.zeros(MAX_ARRAY_LENGTH, dtype=float))


class Plot(Module):
    name = 'plot'
    StateClass = PlotState

    def advance(self, state: State, previous_time: float):
        step_num = state.plot.step_num

        state.plot.time_steps[step_num] = state.time
        state.plot.fungal_burdens[step_num] = len(state.fungus.cells.alive())
        state.plot.macrophage_counts[step_num] = len(state.macrophage.cells.alive())
        state.plot.neutrophil_counts[step_num] = len(state.neutrophil.cells.alive())
        # define additional values to be plotted here

        state.plot.step_num += 1

        return state

    def finalize(self, state: State):
        num_steps = state.plot.step_num
        time_steps = state.plot.time_steps[:num_steps]
        fungal_burdens = state.plot.fungal_burdens[:num_steps]
        macrophage_counts = state.plot.macrophage_counts[:num_steps]
        neutrophil_counts = state.plot.neutrophil_counts[:num_steps]
        # define additional y_val here

        plt.figure()
        plt.xlabel('Time')
        plt.grid(True)

        for y_val, y_label, title, filename in [
            (fungal_burdens, 'Fungal Burden', 'Fungal Burden vs Time', 'fungal_burden_vs_time'),
            (
                macrophage_counts,
                'Macrophage Count',
                'Macrophage Count vs Time',
                'macrophage_count_vs_time',
            ),
            (
                neutrophil_counts,
                'Neutrophil Count',
                'Neutrophil Count vs Time',
                'neutrophil_count_vs_time',
            ),
            # define additional plots here at 4-tuples
        ]:
            plt.clf()
            plt.plot(time_steps, y_val)
            plt.ylabel(y_label)
            plt.title(title)
            plt.savefig(f'{filename }.png')

        return state



from collections.abc import MutableMapping
from enum import IntEnum
from typing import Iterable, Iterator, List, Optional, Union

import attr
from h5py import Group
import numpy as np
from scipy.sparse import coo_matrix, dok_matrix as sparse_matrix
from scipy.spatial.transform import Rotation

from nlisim.cell import CellData, CellList, CellType
from nlisim.coordinates import Point
from nlisim.grid import RectangularGrid
from nlisim.module import Module, ModuleState
from nlisim.random import rg
from nlisim.state import get_class_path, State


class AfumigatusCellData(CellData):
    GROWTH_SCALE_FACTOR = 0.02  # from original code
    BOOLEAN_NETWORK_LENGTH = 23

    class Status(IntEnum):
        RESTING_CONIDIA = 0
        SWELLING_CONIDIA = 1
        GERMTUBE = 5
        HYPHAE = 2
        DYING = 3
        DEAD = 4
        INTERNALIZED = 6

    class State(IntEnum):
        FREE = 0
        INTERNALIZING = 1
        RELEASING = 2

    AFUMIGATUS_FIELDS = [
        ('boolean_network', 'b1', BOOLEAN_NETWORK_LENGTH),
        ('growth', Point.dtype),
        ('growable', 'b1'),
        ('switched', 'b1'),
        ('branchable', 'b1'),
        ('state', 'u1'),
        ('status', 'u1'),
        ('iron_pool', 'f8'),
        ('iron', 'b1'),
        ('iteration', 'i4'),
        ('health', 'f8'),
        ('mobile', 'b1'),
    ]

    FIELDS = CellData.FIELDS + AFUMIGATUS_FIELDS
    dtype = np.dtype(FIELDS, align=True)  # type: ignore

    @classmethod
    def create_cell_tuple(
        cls,
        *,
        iron_pool: float = 0,
        status: Status = Status.RESTING_CONIDIA,
        state: State = State.FREE,
        **kwargs,
    ) -> np.record:

        growth = cls.GROWTH_SCALE_FACTOR * Point.from_array(2 * rg.random(3) - 1)
        network = cls.initial_boolean_network()
        growable = True
        switched = False
        branchable = False
        iteration = 0
        iron = False
        health = 100
        mobile = True

        return CellData.create_cell_tuple(**kwargs) + (
            network,
            growth,
            growable,
            switched,
            branchable,
            state,
            status,
            iron_pool,
            iron,
            iteration,
            health,
            mobile,
        )

    @classmethod
    def initial_boolean_network(cls) -> np.ndarray:
        return np.asarray(
            [
                True,
                False,
                True,
                False,
                True,
                True,
                True,
                True,
                True,
                False,
                False,
                False,
                True,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
            ]
        )


@attr.s(auto_attribs=True, kw_only=True, frozen=True)
class AfumigatusCell(MutableMapping):
    tree: 'AfumigatusCellTreeList'
    index: int = attr.ib()

    @index.validator
    def _validate_index(self, attribute, value):
        if value >= len(self.tree):
            raise ValueError('Invalid cell index')

    def __attrs_post_init__(self):
        if self.index < 0:
            object.__setattr__(self, 'index', self.index + len(self.tree))

    @property
    def record(self) -> CellType:
        return self.tree.cells[self.index]

    @property
    def parent(self) -> Optional['AfumigatusCell']:
        indices = self.tree.adjacency[:, self.index].nonzero()[0]
        if len(indices) == 1:
            return None
        if len(indices) == 2 and indices[0] == self.index:
            index = indices[0]
            if index == self.index:
                index = indices[1]
            return self.__class__(tree=self.tree, index=index)

        raise Exception(f'Invalid adjacency matrix at index {self.index}')

    @property
    def children(self) -> Iterator['AfumigatusCell']:
        indices = self.tree.adjacency[self.index].nonzero()[1]
        for index in indices:
            if index != self.index:
                yield self.__class__(tree=self.tree, index=index)

    def add_child(self, cell: Union[CellType, 'AfumigatusCell']):
        if isinstance(cell, AfumigatusCell):
            cell = cell.record
        self.tree.append(cell, parent=self.index)

    def kill(self) -> List['AfumigatusCell']:
        """Mark the current cell as dead and remove it from the tree."""
        return self.tree.kill(self.index)

    def __getitem__(self, key):
        return self.record.__getitem__(key)

    def __setitem__(self, key, value):
        self.tree.cell_data[key][self.index] = value

    def __delitem__(self, key):
        return self.record.__delitem__(key)

    def __iter__(self):
        for key in self.record.dtype.fields.keys():
            yield key

    def __contains__(self, key):
        return key in self.record.dtype.fields

    def __len__(self):
        return len(self.record)


@attr.s(kw_only=True, frozen=True, repr=False)
class AfumigatusCellList(CellList):
    CellDataClass = AfumigatusCellData


@attr.s(kw_only=True, frozen=True, repr=False)
class AfumigatusCellTreeList(object):
    cells: CellList = attr.ib()
    _adjacency: sparse_matrix = attr.ib()

    @_adjacency.default
    def __set_default_adjacency(self):
        return sparse_matrix((0, 0))

    def __attrs_post_init__(self):
        cells = self.cells
        adjacency = self._adjacency

        object.__setattr__(self, '_adjacency', sparse_matrix((cells.max_cells, cells.max_cells)))

        if len(cells) > 0:
            for key, value in adjacency.items():
                self._adjacency[key] = value

    @property
    def adjacency(self) -> sparse_matrix:
        return self._adjacency[: len(self), : len(self)]

    @property
    def roots(self) -> List[CellType]:
        # This is linear in the number of nodes... might need to be optimized.
        roots = []
        for i in range(len(self)):
            cell = self[i]
            if not cell['dead'] and cell.parent is None:
                roots.append(cell)
        return roots

    @property
    def cell_data(self) -> AfumigatusCellData:
        return self.cells.cell_data

    @property
    def max_cells(self) -> int:
        return self.cells.max_cells

    @property
    def grid(self) -> RectangularGrid:
        return self.cells.grid

    @classmethod
    def random_branch_direction(cls, growth: np.ndarray) -> np.ndarray:
        """Rotate by a random 45 degree angle from the axis of growth."""
        growth = Point.from_array(growth)

        # get a random vector orthogonal to the growth vector
        axis = Point.from_array(np.cross(growth, rg.standard_normal(3)))
        axis = (np.pi / 4) * axis / axis.norm()

        # rotate the growth vector 45 degrees along the random axis
        rotation = Rotation.from_rotvec(axis)
        return rotation.apply(growth)

    @classmethod
    def create_from_seed(cls, grid, **kwargs) -> 'AfumigatusCellTreeList':
        cells = AfumigatusCellList.create_from_seed(grid, **kwargs)
        adjacency = sparse_matrix((1, 1))
        adjacency[0, 0] = 1

        return cls(cells=cells, adjacency=adjacency)

    def __len__(self):
        return len(self.cells)

    def __getitem__(self, index: int) -> AfumigatusCell:
        if isinstance(index, str):
            raise TypeError('Expected an integer index, did you mean `tree.cell_data[key]`?')
        return AfumigatusCell(tree=self, index=index)

    def append(self, cell, parent=None):
        index = len(self)
        self.cells.append(cell)
        self._adjacency[index, index] = 1

        if parent is not None:
            self._adjacency[parent, index] = 1
            iron_pool = self.cells[parent]['iron_pool'] / 2
            self.cells[parent]['iron_pool'] = self.cells[-1]['iron_pool'] = iron_pool

    def extend(self, cells: CellData, parents: Iterable[Union[int, None]] = None) -> None:
        if parents is None:
            parents = [None] * len(cells)

        for cell, parent in zip(cells, parents):
            self.append(cell, parent)

    def kill(self, index: int) -> List[AfumigatusCell]:
        """Kill the cell at the given index and remove it from the tree.

        This method will return a list of cells which were children of the
        killed cell.  These children will be the roots of newly formed trees.
        """
        self.cell_data[index]['dead'] = True
        roots = []
        for i in range(len(self)):
            if i == index:
                continue
            if self._adjacency.pop((i, index), None):
                self[i]['growable'] = True
            if self._adjacency.pop((index, i), None):
                roots.append(self[i])
        return roots

    def is_growable(self) -> np.ndarray:
        cells = self.cells.cell_data
        return self.cells.alive(
            cells['growable']
            & cells.point_mask(cells['point'] + cells['growth'], self.grid)
            & (cells['status'] == AfumigatusCellData.Status.HYPHAE)
        )

    def is_branchable(self, branch_probability: float) -> np.ndarray:
        cells = self.cells.cell_data
        return self.cells.alive(
            cells['branchable']
            & (rg.random(cells.shape) < branch_probability)
            & (cells['status'] == AfumigatusCellData.Status.HYPHAE)
        )

    def absorb_iron(self):
        """Absorb iron from the environment.

        This is a dummy method for the visualization testing purpose.
        """
        #  TODO: implement a real iron uptake model
        cells = self.cells.cell_data
        iron = rg.random(len(self.cells))
        cells['iron_pool'] = np.add(cells['iron_pool'], iron)

    def age(self):
        """Add 1 to iteration each timestep."""
        cells = self.cells.cell_data
        cells['iteration'] += 1

    def alive(self):
        return self.cells.alive()

    def elongate(self):
        cells = self.cell_data
        mask = self.is_growable()
        if len(mask) == 0:
            return

        cells['growable'][mask] = False
        cells['branchable'][mask] = True

        children = AfumigatusCellData(len(mask), initialize=True)
        children['point'] = cells['point'][mask] + cells['growth'][mask]
        children['growth'] = cells['growth'][mask]

        self.extend(children, parents=mask)

    def branch(self, branch_probability: float):
        cells = self.cell_data
        indices = self.is_branchable(branch_probability)
        if len(indices) == 0:
            return

        children = AfumigatusCellData(len(indices), initialize=True)
        children['growth'] = cells[indices]['growth']
        children['growth'] = np.apply_along_axis(
            self.random_branch_direction, 1, children['growth']
        )

        children['point'] = cells['point'][indices] + children['growth']

        # filter out children lying outside of the domain
        indices = indices[cells.point_mask(children['point'], self.grid)]
        if len(indices) == 0:
            return

        # set properties
        cells['branchable'][indices] = False
        children['growable'] = True
        children['branchable'] = False

        self.extend(children, parents=indices)

    def save(self, group: Group, name: str, metadata: dict) -> Group:
        composite_group = group.create_group(name)
        composite_group.attrs['class'] = get_class_path(self)
        composite_group.attrs['type'] = 'CellTree'
        self.cells.save(composite_group, 'cells', metadata)

        sp = self.adjacency.tocoo()
        composite_group.create_dataset(name='row', data=sp.row)
        composite_group.create_dataset(name='col', data=sp.col)
        composite_group.create_dataset(name='value', data=sp.data)
        return composite_group

    @classmethod
    def load(
        cls, global_state: State, group: Group, name: str, metadata: dict
    ) -> 'AfumigatusCellTreeList':
        composite_dataset = group[name]
        cells = AfumigatusCellList.load(global_state, composite_dataset, 'cells', metadata)

        adjacency = coo_matrix(
            (
                composite_dataset['value'],
                (composite_dataset['row'], composite_dataset['col']),
            ),
            shape=(cells.max_cells, cells.max_cells),
        ).todok()

        return cls(cells=cells, adjacency=adjacency)


def cell_list_factory(self: 'AfumigatusState'):
    cells = AfumigatusCellList(grid=self.global_state.grid)
    return AfumigatusCellTreeList(cells=cells)


@attr.s(kw_only=True)
class AfumigatusState(ModuleState):
    tree: AfumigatusCellTreeList = attr.ib(default=attr.Factory(cell_list_factory, takes_self=True))
    p_lodge: float = 0.5
    ITER_TO_CHANGE_STATUS: int = 2


class Afumigatus(Module):
    name = 'afumigatus'
    StateClass = AfumigatusState

    def initialize(self, state: State):
        afumigatus: AfumigatusState = state.afumigatus

        afumigatus.p_lodge = self.config.getfloat('p_lodge')
        afumigatus.ITER_TO_CHANGE_STATUS = self.config.getint('ITER_TO_CHANGE_STATUS')

        point = Point(x=4, y=4, z=4)
        afumigatus.tree = AfumigatusCellTreeList.create_from_seed(
            grid=state.grid, point=point, status=AfumigatusCellData.Status.HYPHAE
        )

        return state

    def advance(self, state: State, previous_time: float) -> State:
        # afumigatus: AfumigatusState = state.afumigatus
        # grid = state.grid
        # tissue = state.geometry.lung_tissue
        #
        # for index in afumigatus.tree.cells.alive():
        #    cell = afumigatus.tree.cells[index]
        #    vox = grid.get_voxel(cell['point'])
        #
        #    if cell['mobile']:
        #        if vox.x > Voxel(x=grid.xv[-1], y=grid.yv[0], z=grid.zv[0]):
        #            cell['status'] = AfumigatusCellData.Status.DEAD
        #            cell['dead'] = True
        #        if tissue[vox.z, vox.y, vox.x] == TissueTypes.EPITHELIUM:
        #            if afumigatus.p_lodge > rg.random():
        #                cell['mobile'] = False
        #    if cell['iteration'] >= afumigatus.ITER_TO_CHANGE_STATUS:
        #        if cell['status'] == AfumigatusCellData.Status.RESTING_CONIDIA:
        #            cell['status'] = AfumigatusCellData.Status.SWELLING_CONIDIA
        #        elif cell['status'] == AfumigatusCellData.Status.SWELLING_CONIDIA:
        #            cell['status'] = AfumigatusCellData.Status.GERMTUBE
        #        elif cell['status'] == AfumigatusCellData.Status.INTERNALIZED:
        #            if afumigatus.p_internal_swell > rg.random():
        #                cell['status'] = AfumigatusCellData.Status.SWELLING_CONIDIA
        #
        #        if cell['status'] == AfumigatusCellData.Status.SWELLING_CONIDIA:
        #            cell['iteration'] = 0

        return state



import numpy as np
from scipy.sparse import csr_matrix, dok_matrix, eye
from scipy.sparse.linalg import cg

from nlisim.coordinates import Voxel
from nlisim.grid import RectangularGrid


def discrete_laplacian(
    grid: RectangularGrid, mask: np.ndarray, dtype: np.dtype = np.float64
) -> csr_matrix:
    """Return a discrete laplacian operator for the given restricted grid.

    This computes a standard laplacian operator as a scipy linear operator, except it is
    restricted to a grid mask.  The use case for this is to compute surface diffusion
    on a gridded variable.  The mask is generated from a category on the lung_tissue
    variable.
    """
    graph_shape = len(grid), len(grid)
    laplacian = dok_matrix(graph_shape)

    delta_z = grid.delta(0)
    delta_y = grid.delta(1)
    delta_x = grid.delta(2)

    for k, j, i in zip(*(mask).nonzero()):
        voxel = Voxel(x=i, y=j, z=k)
        voxel_index = grid.get_flattened_index(voxel)
        normalization = 0

        for neighbor in grid.get_adjecent_voxels(voxel, corners=False):
            ni = neighbor.x
            nj = neighbor.y
            nk = neighbor.z

            if not mask[nk, nj, ni]:
                continue

            neighbor_index = grid.get_flattened_index(neighbor)

            dx = delta_x[k, j, i] * (i - ni)
            dy = delta_y[k, j, i] * (j - nj)
            dz = delta_z[k, j, i] * (k - nk)
            distance2 = 1 / (dx * dx + dy * dy + dz * dz)

            normalization -= distance2
            laplacian[voxel_index, neighbor_index] = distance2

        laplacian[voxel_index, voxel_index] = normalization

    return laplacian.tocsr()


def apply_diffusion(
    variable: np.ndarray, laplacian: csr_matrix, diffusivity: float, dt: float
) -> np.ndarray:
    """Apply diffusion to a variable.

    Solves laplaces equation in 3D using implicit time steps.  The variable is
    advanced in time by `dt` time units using GMRES.

    The intended use case for this method is to perform "surface diffusion" generated
    by a mask from the `lung_tissue` variable, e.g.

        surface_mask = lung_tissue == TissueTypes.EPITHELIUM
        laplacian = discrete_laplacian(grid, mask)
        iron_concentration[:] = apply_diffusion(iron_concentration, laplacian, diffusivity, dt)
    """
    operator = eye(*laplacian.shape) - (diffusivity * dt) * laplacian
    var_next, info = cg(operator, variable.ravel())
    if info != 0:
        raise Exception(f'GMRES failed ({info})')
    return var_next.reshape(variable.shape)



from io import BytesIO

from h5py import File
import numpy as np
from pytest import fixture

from nlisim.config import SimulationConfig
from nlisim.coordinates import Point
from nlisim.grid import RectangularGrid
from nlisim.state import State


@fixture
def base_config():
    yield SimulationConfig()


@fixture
def config():
    yield SimulationConfig({'simulation': {'modules': 'nlisim.modules.afumigatus.Afumigatus'}})


@fixture
def state(config):
    yield State.create(config)


@fixture
def grid():
    # a 100 x 100 x 100 unit grid
    yield RectangularGrid.construct_uniform((10, 10, 10), (10, 10, 10))


@fixture
def point():
    yield Point(x=50, y=50, z=50)


@fixture
def hdf5_file():
    yield File(BytesIO(), 'w')


@fixture
def hdf5_group(hdf5_file):
    yield hdf5_file.create_group('test-group')


@fixture
def epi_geometry():
    # a 10 x 10 x 10 grid with epithelium
    tissue = np.empty((10, 10, 10))
    tissue.fill(3)
    yield tissue


@fixture
def air_geometry():
    # a 10 x 10 x 10 grid with air
    tissue = np.empty((10, 10, 10))
    tissue.fill(0)
    yield tissue



from contextlib import contextmanager
from typing import Callable, Iterator, List, TYPE_CHECKING


if TYPE_CHECKING:  # prevent circular imports for type checking
    from nlisim.state import State  # noqa

ValidatorMethod = Callable[['State'], None]


class ValidationError(Exception):
    """
    An error type raised when a simulation validation condition is violated.

    This also provides the ability to report the context of a validation
    error... that is, to report the specific step causing the validation error.
    """

    def __init__(self, msg: str):
        super().__init__(msg)
        self._ctx: List['str'] = []

    def push_context(self, ctx: str) -> None:
        """
        Push a new execution context onto the exception.

        The context is an arbitrary string meant to provide users a hint as to
        the cause of an error even when plugins are allowed to make arbitrary
        changes to the system state.
        """
        self._ctx.append(ctx)

    def __str__(self) -> str:
        msg = super().__str__()
        for ctx in self._ctx:
            msg = f'After execution of "{ctx}": ' + msg
        return msg


@contextmanager
def context(name: str) -> Iterator[None]:
    """Create a new "validation context".

    This returns a context manager.  Any validation error thrown within this context
    will contain a reference to the value provided.
    """
    try:
        yield
    except ValidationError as e:
        e.push_context(name)
        raise
    except Exception:
        # TODO: use simulation logger
        print(f'ERROR: Unhandled exception raised while executing "{name}"')
        raise



#!/usr/bin/env python
import os

from setuptools import find_packages, setup


def prerelease_local_scheme(version):
    """Return local scheme version unless building on master in Gitlab.

    This function returns the local scheme version number
    (e.g. 0.0.0.dev<N>+g<HASH>) unless building on Gitlab for a
    pre-release in which case it ignores the hash and produces a
    PEP440 compliant pre-release version number (e.g. 0.0.0.dev<N>).
    """
    from setuptools_scm.version import get_local_node_and_date

    if 'CI_COMMIT_REF_NAME' in os.environ and os.environ['CI_COMMIT_REF_NAME'] == 'master':
        return ''
    else:
        return get_local_node_and_date(version)


setup(
    name='nlisim',
    packages=find_packages(exclude=['test', 'test.*']),
    install_requires=[
        'attrs',
        'click',
        'click-pathlib',
        'h5py',
        'matplotlib',
        'numpy',
        'scipy',
        'tqdm',
        'vtk',
    ],
    entry_points={'console_scripts': ['nlisim = nlisim.cli:main']},
    setup_requires=['setuptools_scm'],
    use_scm_version={'local_scheme': prerelease_local_scheme},
)



import attr
import numpy as np

from nlisim.cell import CellData, CellList
from nlisim.coordinates import Point, Voxel
from nlisim.grid import RectangularGrid
from nlisim.module import Module, ModuleState
from nlisim.modules.fungus import FungusCellData, FungusCellList
from nlisim.modules.geometry import TissueTypes
from nlisim.random import rg
from nlisim.state import State

MAX_CONIDIA = 100


class MacrophageCellData(CellData):

    MACROPHAGE_FIELDS = [
        ('iteration', 'i4'),
        ('phagosome', (np.int32, (MAX_CONIDIA))),
    ]

    dtype = np.dtype(CellData.FIELDS + MACROPHAGE_FIELDS, align=True)  # type: ignore

    @classmethod
    def create_cell_tuple(
        cls,
        **kwargs,
    ) -> np.record:

        iteration = 0
        phagosome = np.empty(MAX_CONIDIA)
        phagosome.fill(-1)
        return CellData.create_cell_tuple(**kwargs) + (
            iteration,
            phagosome,
        )


@attr.s(kw_only=True, frozen=True, repr=False)
class MacrophageCellList(CellList):
    CellDataClass = MacrophageCellData

    def len_phagosome(self, index):
        cell = self[index]
        return len(np.argwhere(cell['phagosome'] != -1))

    def append_to_phagosome(self, index, pathogen_index, max_size):
        cell = self[index]
        index_to_append = MacrophageCellList.len_phagosome(self, index)
        if (
            index_to_append < MAX_CONIDIA
            and index_to_append < max_size
            and pathogen_index not in cell['phagosome']
        ):
            cell['phagosome'][index_to_append] = pathogen_index
            return True
        else:
            return False

    def remove_from_phagosome(self, index, pathogen_index):
        phagosome = self[index]['phagosome']
        if pathogen_index in phagosome:
            itemindex = np.argwhere(phagosome == pathogen_index)[0][0]
            size = MacrophageCellList.len_phagosome(self, index)
            if itemindex == size - 1:
                # full phagosome
                phagosome[itemindex] = -1
                return True
            else:
                phagosome[itemindex:-1] = phagosome[itemindex + 1 :]
                phagosome[-1] = -1
                return True
        else:
            return False

    def clear_all_phagosome(self, index, fungus: FungusCellList):
        for i in range(0, self.len_phagosome(index)):
            index = self[index]['phagosome'][i]
            fungus[index]['internalized'] = False
        self[index]['phagosome'].fill(-1)

    def recruit_new(self, rec_rate_ph, rec_r, p_rec_r, tissue, grid, cyto):
        num_reps = rec_rate_ph  # number of macrophages recruited per time step

        blood_index = np.argwhere(tissue == TissueTypes.BLOOD.value)
        blood_index = np.transpose(blood_index)
        mask = cyto[blood_index[2], blood_index[1], blood_index[0]] >= rec_r
        blood_index = np.transpose(blood_index)
        cyto_index = blood_index[mask]
        rg.shuffle(cyto_index)

        for _ in range(0, num_reps):
            if len(cyto_index) > 0:
                ii = rg.integers(len(cyto_index))
                point = Point(
                    x=grid.x[cyto_index[ii][2]],
                    y=grid.y[cyto_index[ii][1]],
                    z=grid.z[cyto_index[ii][0]],
                )

                if p_rec_r > rg.random():
                    self.append(MacrophageCellData.create_cell(point=point))

    def absorb_cytokines(self, m_abs, cyto, grid):
        for index in self.alive():
            vox = grid.get_voxel(self[index]['point'])
            x = vox.x
            y = vox.y
            z = vox.z
            cyto[z, y, x] = (1 - m_abs) * cyto[z, y, x]

    def produce_cytokines(self, m_det, m_n, grid, fungus: FungusCellList, cyto):
        for i in self.alive():
            vox = grid.get_voxel(self[i]['point'])

            hyphae_count = 0

            x_r = []
            y_r = []
            z_r = []

            if m_det == 0:
                index_arr = fungus.get_cells_in_voxel(vox)
                for index in index_arr:
                    if fungus[index]['form'] == FungusCellData.Form.HYPHAE:
                        hyphae_count += 1

            else:
                for num in range(0, m_det + 1):
                    x_r.append(num)
                    y_r.append(num)
                    z_r.append(num)

                for num in range(-1 * m_det, 0):
                    x_r.append(num)
                    y_r.append(num)
                    z_r.append(num)

                for x in x_r:
                    for y in y_r:
                        for z in z_r:
                            zk = vox.z + z
                            yj = vox.y + y
                            xi = vox.x + x
                            if grid.is_valid_voxel(Voxel(x=xi, y=yj, z=zk)):
                                index_arr = fungus.get_cells_in_voxel(Voxel(x=xi, y=yj, z=zk))
                                for index in index_arr:
                                    if fungus[index]['form'] == FungusCellData.Form.HYPHAE:
                                        hyphae_count += 1

            cyto[vox.z, vox.y, vox.x] = cyto[vox.z, vox.y, vox.x] + m_n * hyphae_count

    def move(self, rec_r, grid, cyto, tissue, fungus: FungusCellList):
        for cell_index in self.alive():
            cell = self[cell_index]
            vox = grid.get_voxel(cell['point'])

            p = np.zeros(shape=27)
            vox_list = []
            i = -1

            for x in [0, 1, -1]:
                for y in [0, 1, -1]:
                    for z in [0, 1, -1]:
                        zk = vox.z + z
                        yj = vox.y + y
                        xi = vox.x + x
                        if (
                            grid.is_valid_voxel(Voxel(x=xi, y=yj, z=zk))
                            and tissue[zk, yj, xi] != TissueTypes.AIR.value
                        ):
                            vox_list.append([x, y, z])
                            i += 1
                            if cyto[zk, yj, xi] >= rec_r:
                                p[i] = cyto[zk, yj, xi]

            indices = np.argwhere(p != 0)
            num_vox_possible = len(indices)
            if num_vox_possible == 1:
                i = indices[0][0]
            elif num_vox_possible > 1:
                inds = np.argwhere(p == p[np.argmax(p)])
                rg.shuffle(inds)
                i = inds[0][0]
            else:
                i = rg.integers(len(vox_list))

            point = Point(
                x=grid.x[vox.x + vox_list[i][0]],
                y=grid.y[vox.y + vox_list[i][1]],
                z=grid.z[vox.z + vox_list[i][2]],
            )

            cell['point'] = point
            self.update_voxel_index([cell_index])

            for i in range(0, self.len_phagosome(cell_index)):
                f_index = cell['phagosome'][i]
                fungus[f_index]['point'] = point
                fungus.update_voxel_index([f_index])

    def internalize_conidia(self, m_det, max_spores, p_in, grid, fungus: FungusCellList):
        for i in self.alive():
            cell = self[i]
            vox = grid.get_voxel(cell['point'])

            x_r = []
            y_r = []
            z_r = []

            if m_det == 0:
                index_arr = fungus.get_cells_in_voxel(vox)
                for index in index_arr:
                    if (
                        fungus[index]['form'] == FungusCellData.Form.CONIDIA
                        and not fungus[index]['internalized']
                        and p_in > rg.random()
                    ):
                        fungus[index]['internalized'] = True
                        self.append_to_phagosome(i, index, max_spores)
            else:
                for num in range(0, m_det + 1):
                    x_r.append(num)
                    y_r.append(num)
                    z_r.append(num)

                for num in range(-1 * m_det, 0):
                    x_r.append(num)
                    y_r.append(num)
                    z_r.append(num)

                for x in x_r:
                    for y in y_r:
                        for z in z_r:
                            zk = vox.z + z
                            yj = vox.y + y
                            xi = vox.x + x
                            if grid.is_valid_voxel(Voxel(x=xi, y=yj, z=zk)):
                                index_arr = fungus.get_cells_in_voxel(Voxel(x=xi, y=yj, z=zk))
                                for index in index_arr:
                                    if (
                                        fungus[index]['form'] == FungusCellData.Form.CONIDIA
                                        and not fungus[index]['internalized']
                                        and p_in > rg.random()
                                    ):
                                        fungus[index]['internalized'] = True
                                        self.append_to_phagosome(i, index, max_spores)

    def damage_conidia(self, kill, t, health, fungus):
        for i in self.alive():
            cell = self[i]
            for ii in range(0, self.len_phagosome(i)):
                index = cell['phagosome'][ii]
                fungus[index]['health'] = fungus[index]['health'] - (health * (t / kill))
                if fungus[index]['dead']:
                    self.remove_from_phagosome(i, index)

    def remove_if_sporeless(self, val):
        living = self.alive()
        living_len = len(living)
        num = int(val * living_len)
        if num == 0 and living_len > 0:
            num = 1
        for _ in range(num):
            r = rg.integers(living_len)
            self.cell_data[living[r]]['dead'] = True


def cell_list_factory(self: 'MacrophageState'):
    return MacrophageCellList(grid=self.global_state.grid)


@attr.s(kw_only=True)
class MacrophageState(ModuleState):
    cells: MacrophageCellList = attr.ib(default=attr.Factory(cell_list_factory, takes_self=True))
    rec_r: float
    p_rec_r: float
    m_abs: float
    m_n: float
    kill: float
    m_det: int
    rec_rate_ph: int
    time_m: float
    max_conidia_in_phag: int
    p_internalization: float
    rm: float


class Macrophage(Module):
    name = 'macrophage'
    defaults = {
        'cells': '',
        'rec_r': '1.0',
        'p_rec_r': '1.0',
        'm_abs': '0.1 ',
        'm_n': '10.0 ',
        'kill': '10.0',
        'm_det': '15',
        'rec_rate_ph': '2',
        'time_m': '1',
        'max_conidia_in_phag': '50',
        'p_internalization': '0.3',
        'rm': '0.05',
    }
    StateClass = MacrophageState

    def initialize(self, state: State):
        macrophage: MacrophageState = state.macrophage
        grid: RectangularGrid = state.grid

        macrophage.rec_r = self.config.getfloat('rec_r')
        macrophage.p_rec_r = self.config.getfloat('p_rec_r')
        macrophage.m_abs = self.config.getfloat('m_abs')
        macrophage.m_n = self.config.getfloat('Mn')
        macrophage.kill = self.config.getfloat('kill')
        macrophage.m_det = self.config.getint('m_det')  # radius
        macrophage.rec_rate_ph = self.config.getint('rec_rate_ph')
        macrophage.time_m = self.config.getfloat('time_step')
        macrophage.max_conidia_in_phag = self.config.getint('max_conidia_in_phag')
        macrophage.rm = self.config.getfloat('rm')
        macrophage.p_internalization = self.config.getfloat('p_internalization')

        macrophage.cells = MacrophageCellList(grid=grid)

        return state

    def advance(self, state: State, previous_time: float):
        macrophage: MacrophageState = state.macrophage
        m_cells: MacrophageCellList = macrophage.cells
        tissue = state.geometry.lung_tissue
        grid = state.grid
        cyto = state.molecules.grid['m_cyto']
        n_cyto = state.molecules.grid['n_cyto']
        fungus: FungusCellList = state.fungus.cells
        health = state.fungus.health

        # recruit new
        m_cells.recruit_new(
            macrophage.rec_rate_ph, macrophage.rec_r, macrophage.p_rec_r, tissue, grid, cyto
        )

        # absorb cytokines
        m_cells.absorb_cytokines(macrophage.m_abs, cyto, grid)

        # produce cytokines
        m_cells.produce_cytokines(macrophage.m_det, macrophage.m_n, grid, fungus, n_cyto)

        # move
        m_cells.move(macrophage.rec_r, grid, cyto, tissue, fungus)

        # internalize
        m_cells.internalize_conidia(
            macrophage.m_det,
            macrophage.max_conidia_in_phag,
            macrophage.p_internalization,
            grid,
            fungus,
        )

        # damage conidia
        m_cells.damage_conidia(macrophage.kill, macrophage.time_m, health, fungus)

        if len(fungus.alive(fungus.cell_data['form'] == FungusCellData.Form.CONIDIA)) == 0:
            m_cells.remove_if_sporeless(macrophage.rm)

        return state



import pytest

from nlisim.coordinates import Point, Voxel
from nlisim.grid import RectangularGrid


def v(x: int, y: int, z: int) -> Voxel:
    return Voxel(x=x, y=y, z=z)


def p(x: float, y: float, z: float) -> Point:
    return Point(x=x, y=y, z=z)


@pytest.fixture
def grid():
    shape = [10, 20, 30]
    spacing = [1, 1, 1]
    yield RectangularGrid.construct_uniform(shape=shape, spacing=spacing)


@pytest.mark.parametrize(
    'point,voxel',
    [
        (p(0.5, 0.5, 0.5), v(0, 0, 0)),
        (p(1.5, 1.5, 1.5), v(1, 1, 1)),
        (p(0.9, 0.9, 0.9), v(0, 0, 0)),
        (p(1.9, 1.1, 0.1), v(1, 1, 0)),
        (p(-0.1, -0.4, 0.4), v(-1, -1, 0)),
    ],
)
def test_get_voxel(grid: RectangularGrid, point, voxel):
    assert grid.get_voxel(point) == voxel


@pytest.mark.parametrize(
    'voxel,valid',
    [
        (v(0, 0, 0), True),
        (v(9, 0, 4), True),
        (v(-1, 0, 0), False),
        (v(25, 0, 0), True),
        (v(0, 0, 25), False),
        (v(-1, -1, 100), False),
    ],
)
def test_valid_voxel(grid: RectangularGrid, voxel, valid):
    assert grid.is_valid_voxel(voxel) == valid


# fmt: off
@pytest.mark.parametrize(
    'voxel,neighbors',
    [
        (
            v(5, 5, 5),
            {
                v(4, 5, 5),
                v(6, 5, 5),
                v(5, 4, 5),
                v(5, 6, 5),
                v(5, 5, 4),
                v(5, 5, 6),
            }
        ),
        (
            v(0, 5, 5),
            {
                v(1, 5, 5),
                v(0, 4, 5),
                v(0, 6, 5),
                v(0, 5, 4),
                v(0, 5, 6),
            }
        ),
        (
            v(0, 0, 5),
            {
                v(1, 0, 5),
                v(0, 1, 5),
                v(0, 0, 4),
                v(0, 0, 6),
            }
        ),
        (
            v(0, 0, 0),
            {
                v(1, 0, 0),
                v(0, 1, 0),
                v(0, 0, 1),
            }
        )
    ]
)
# fmt: on
def test_get_adjacent_voxels(grid: RectangularGrid, voxel, neighbors):
    assert set(grid.get_adjecent_voxels(voxel)) == neighbors


# fmt: off
@pytest.mark.parametrize(
    'voxel,neighbors',
    [
        (
            v(5, 5, 5),
            26
        ),
        (
            v(0, 5, 5),
            17
        ),
        (
            v(0, 0, 5),
            11
        ),
        (
            v(0, 0, 0),
            7
        )
    ]
)
# fmt: on
def test_get_corner_adjacent_voxels(grid: RectangularGrid, voxel, neighbors):
    assert len(list(grid.get_adjecent_voxels(voxel, corners=True))) == neighbors


@pytest.mark.parametrize(
    'point,in_domain',
    [
        (p(0.1, 0.1, 0.1), True),
        (p(0, 0, 0), True),
        (p(-1, 0, 0), False),
        (p(100, 0, 0), False),
        (p(100, 10, -100), False),
    ],
)
def test_point_in_domain(grid, point, in_domain):
    assert grid.is_point_in_domain(point) == in_domain


@pytest.mark.parametrize(
    'point,voxel',
    [
        (p(0.5, 0.5, 0.5), v(0, 0, 0)),
        (p(1.5, 0.5, 0.5), v(1, 0, 0)),
        (p(1.1, 5.5, 1.9), v(1, 5, 1)),
    ],
)
def test_get_nearest_voxel(grid, point, voxel):
    assert grid.get_nearest_voxel(point) == voxel


# fmt: off
@pytest.mark.parametrize(
    'point,distance,voxels',
    [
        (
            p(5.5, 5.5, 5.5),
            1,
            {
                (v(5, 5, 5), 0),
                (v(4, 5, 5), 1),
                (v(6, 5, 5), 1),
                (v(5, 4, 5), 1),
                (v(5, 6, 5), 1),
                (v(5, 5, 4), 1),
                (v(5, 5, 6), 1),
            }
        ),
        (
            p(0.5, 5.5, 5.5),
            1,
            {
                (v(0, 5, 5), 0),
                (v(1, 5, 5), 1),
                (v(0, 4, 5), 1),
                (v(0, 6, 5), 1),
                (v(0, 5, 4), 1),
                (v(0, 5, 6), 1),
            }
        ),
        (
            p(0.5, 0.5, 5.5),
            1,
            {
                (v(0, 0, 5), 0),
                (v(1, 0, 5), 1),
                (v(0, 1, 5), 1),
                (v(0, 0, 4), 1),
                (v(0, 0, 6), 1),
            }
        ),
        (
            p(0.5, 0.5, 0.5),
            1,
            {
                (v(0, 0, 0), 0),
                (v(1, 0, 0), 1),
                (v(0, 1, 0), 1),
                (v(0, 0, 1), 1),
            }
        )
    ]
)
# fmt: on
def test_get_voxels_in_range(grid: RectangularGrid, point, distance, voxels):
    assert set(grid.get_voxels_in_range(point, distance)) == voxels


def test_get_voxels_in_large_range(grid: RectangularGrid):
    for voxel, distance in grid.get_voxels_in_range(Point(x=4, y=4, z=4), 10):
        assert distance <= 10
        assert grid.is_valid_voxel(voxel)


@pytest.mark.parametrize(
    'voxel,index',
    [(v(0, 0, 0), 0), (v(1, 0, 0), 1), (v(0, 1, 0), 30), (v(0, 0, 1), 600), (v(1, 1, 1), 631)],
)
def test_get_flattened_index(voxel, index, grid: RectangularGrid):
    assert grid.get_flattened_index(voxel) == index
    assert grid.voxel_from_flattened_index(index) == voxel



from h5py import Group
import numpy as np
from numpy.testing import assert_array_equal
from pytest import fixture, raises

from nlisim.cell import CellData, CellList
from nlisim.coordinates import Point, Voxel
from nlisim.grid import RectangularGrid


@fixture
def cell(grid: RectangularGrid, point: Point):
    # a single cell in the middle of the domain
    cell = CellData.create_cell(point=point)
    cells = CellData([cell])
    yield cells


@fixture
def cell_list(grid: RectangularGrid, point: Point):
    cells = CellList.create_from_seed(grid=grid, point=point)
    yield cells


def test_append_cell_list(cell, cell_list):
    original_length = len(cell_list)
    cell_list.append(cell)
    assert len(cell_list) == original_length + 1
    assert cell == cell_list[-1]


def test_extend_cell_list(cell, cell_list):
    original_length = len(cell_list)
    cell_list.extend([cell, cell])
    assert len(cell_list) == original_length + 2
    assert cell == cell_list[-1]
    assert cell == cell_list[-2]


def test_serialize(cell_list: CellList, hdf5_group: Group):
    cell_list_group = cell_list.save(hdf5_group, 'test', {})
    assert cell_list_group['cell_data'].shape == (len(cell_list),)
    assert cell_list_group['cell_data'].dtype == cell_list.cell_data.dtype


def test_get_cell(cell_list: CellList):
    assert cell_list[0] == cell_list.cell_data[0]


def test_getitem_error(cell_list: CellList):
    with raises(TypeError):
        _ = cell_list['a']  # type: ignore


def test_out_of_memory_error(grid: RectangularGrid, cell: CellData):
    cell_list = CellList(grid=grid, max_cells=1)
    cell_list.append(cell)

    with raises(Exception):
        cell_list.append(cell)


def test_filter_out_dead(grid: RectangularGrid):
    cells = CellList(grid=grid)
    cells.extend([CellData.create_cell(dead=bool(i % 2)) for i in range(10)])

    assert_array_equal(cells.alive(), [0, 2, 4, 6, 8])
    assert_array_equal(cells.alive([1, 2, 3, 6]), [2, 6])

    mask = np.arange(10) < 5
    assert_array_equal(cells.alive(mask), [0, 2, 4])


def test_get_neighboring_cells(grid: RectangularGrid):
    point = Point(x=4.5, y=4.5, z=4.5)

    raw_cells = [CellData.create_cell(point=point) for _ in range(5)]
    raw_cells[1]['point'] = Point(x=-1, y=4.5, z=4.5)
    raw_cells[4]['point'] = Point(x=4.5, y=4.5, z=-1)

    cells = CellList(grid=grid)
    cells.extend(raw_cells)

    assert_array_equal(cells.get_neighboring_cells(cells[0]), [0, 2, 3])
    assert_array_equal(cells.get_cells_in_voxel(Voxel(x=0, y=0, z=0)), [0, 2, 3])


def test_move_cell(grid: RectangularGrid):
    point = Point(x=4.5, y=4.5, z=4.5)

    raw_cells = [CellData.create_cell(point=point) for _ in range(5)]
    raw_cells[1]['point'] = Point(x=-1, y=4.5, z=4.5)
    raw_cells[4]['point'] = Point(x=4.5, y=4.5, z=-1)

    cells = CellList(grid=grid)
    cells.extend(raw_cells)

    cells[0]['point'] = Point(x=50, y=50, z=50)

    # updating an incorrect index will not update the cell at index 0
    cells.update_voxel_index([1, 3])
    assert_array_equal(cells.get_neighboring_cells(cells[2]), [0, 2, 3])
    assert cells._reverse_voxel_index[0] == grid.get_voxel(point)

    # this should correctly update the voxel index
    cells.update_voxel_index([0])
    assert_array_equal(cells.get_neighboring_cells(cells[0]), [0])
    assert cells._reverse_voxel_index[0] == grid.get_voxel(cells[0]['point'])



from importlib import import_module
from typing import Any, Dict, Optional, Type, Union

import attr
from h5py import Dataset, Group
import numpy as np

from nlisim.config import SimulationConfig
from nlisim.state import State

AttrValue = Union[float, str, bool, np.ndarray]


@attr.s(auto_attribs=True, kw_only=True)
class ModuleState(object):
    """
    Base type intended to store the state for simulation modules.

    This class contains serialization support for basic types (float, int, str,
    bool) and numpy arrays of those types.  Modules containing more complicated
    state must override the serialization mechanism with custom behavior.
    """

    global_state: 'State'

    def save_state(self, group: Group) -> None:
        """Save the module state into an HDF5 group."""
        for field in attr.fields(type(self)):
            name = field.name
            if name == 'global_state':
                continue
            value = getattr(self, name)
            self.save_attribute(group, name, value, field.metadata)

    @classmethod
    def load_state(cls, global_state: 'State', group: Group) -> 'ModuleState':
        """Load this module's state from an HDF5 group."""
        kwargs: Dict[str, Any] = {'global_state': global_state}
        for field in attr.fields(cls):
            name = field.name
            if name == 'global_state':
                continue
            metadata = field.metadata or {}

            group_object = group.get(name, None)
            if group_object is None:
                raise ValueError(f'Could not read {name} from file.')

            if isinstance(group_object, Group):
                # TODO: break this out into helper methods so subclasses can customize
                class_name = group_object.attrs.get('class')
                if class_name is None:
                    raise ValueError(f'Field {name} contained an invalid composite type.')

                module_name, class_name = class_name.split(':')
                try:
                    module = import_module(module_name)
                except ImportError:
                    raise TypeError(f'File references unknown module {module_name} in {name}')

                class_ = getattr(module, class_name, None)
                if class_ is None:
                    raise TypeError(f'File references invalid class for {name}')

                kwargs[name] = class_.load(global_state, group, name, metadata)

            else:
                kwargs[name] = cls.load_attribute(global_state, group, name, metadata)
        return cls(**kwargs)

    @classmethod
    def save_attribute(
        cls, group: Group, name: str, value: AttrValue, metadata: dict
    ) -> Union[Dataset, Group]:
        """Save an attribute into an HDF5 group."""
        metadata = metadata or {}
        if isinstance(value, (float, int, str, bool, np.ndarray)):
            return cls.save_simple_type(group, name, value, metadata)
        elif hasattr(value, 'save'):
            return value.save(group, name, metadata)
        else:
            # modules must define serializers for unsupported types
            raise TypeError(f'Attribute {name} in {group.name} contains an unsupported datatype.')

    @classmethod
    def save_simple_type(cls, group: Group, name: str, value: AttrValue, metadata: dict) -> Dataset:
        kwargs: Dict[str, Any] = {}
        if metadata.get('grid'):
            kwargs = dict(
                compression='gzip',  # transparent compression
                shuffle=True,  # improve compressiblity
                fletcher32=True,  # checksum
            )

        var = group.create_dataset(name=name, data=np.asarray(value), **kwargs)

        # mark the original attribute as a scalar to unwrap the numpy array when loading
        var.attrs['scalar'] = not isinstance(value, np.ndarray)

        if metadata.get('grid'):
            # attach grid scales to dataset dimensions
            var.dims[0].attach_scale(var.file['z'])
            var.dims[1].attach_scale(var.file['y'])
            var.dims[2].attach_scale(var.file['x'])

        return var

    @classmethod
    def load_attribute(
        cls, global_state: State, group: Group, name: str, metadata: Optional[dict] = None
    ) -> AttrValue:
        """Load a raw value from an HDF5 file group."""
        dataset = group[name]
        if dataset.attrs.get('scalar', False):
            value = dataset[()]
        else:
            value = dataset[:]
        return value


class Module(object):
    name: str = ''
    """A unique name for this module used for namespacing"""

    defaults: Dict[str, str] = {}
    """Default values for all config options"""

    StateClass: Type[ModuleState] = ModuleState
    """Container for extra state required by this module."""

    def __init__(self, config: SimulationConfig):
        if not config.has_section(self.section):
            config.add_section(self.section)

        self.config = config[self.section]
        values = dict(self.defaults, **self.config)
        self.config.update(values)

    @property
    def section(self):
        """Return the section in the configuration object used by this module."""
        return self.name

    # The following are no-op hooks that a module can define to customize the
    # behavior.  A module can override any of these methods to execute
    # arbitrary code during the simulation lifetime.
    def construct(self, state: State) -> None:
        """Run after state construction."""

    def initialize(self, state: State) -> State:
        """Run after state initialization."""
        return state

    def advance(self, state: State, previous_time: float) -> State:
        """Run after advancing the simulation state in time."""
        return state

    def finalize(self, state: State) -> State:
        """Run after the last timestep."""
        return state
